@@ -91,25 +91,25 @@ function loadBreezeScrollBars() {
     // "Failed to read the 'cssRules' property from 'CSSStyleSheet': Cannot access rules"
     var styleTag = document.createElement("style");
     styleTag.appendChild(document.createTextNode(`
-::-webkit-scrollbar {
+html::-webkit-scrollbar {
     /* we'll add padding as "border" in the thumb*/
     height: 20px;
     width: 20px;
     background: white;
 }
 
-::-webkit-scrollbar-track {
+html::-webkit-scrollbar-track {
     border-radius: 20px;
     border: 7px solid white; /* FIXME why doesn't "transparent" work here?! */
     background-color: white;
     width: 6px !important; /* 20px scrollbar - 2 * 7px border */
     box-sizing: content-box;
 }
-::-webkit-scrollbar-track:hover {
+html::-webkit-scrollbar-track:hover {
     background-color: #BFC0C2;
 }
 
-::-webkit-scrollbar-thumb {
+html::-webkit-scrollbar-thumb {
     background-color: #3DAEE9; /* default blue breeze color */
     border: 7px solid transparent;
     border-radius: 20px;
@@ -118,14 +118,14 @@ function loadBreezeScrollBars() {
     box-sizing: content-box;
     min-height: 30px;
 }
-::-webkit-scrollbar-thumb:window-inactive {
+html::-webkit-scrollbar-thumb:window-inactive {
    background-color: #949699; /* when window is inactive it's gray */
 }
-::-webkit-scrollbar-thumb:hover {
+html::-webkit-scrollbar-thumb:hover {
     background-color: #93CEE9; /* hovered is a lighter blue */
 }
 
-::-webkit-scrollbar-corner {
+html::-webkit-scrollbar-corner {
     background-color: white; /* FIXME why doesn't "transparent" work here?! */
 }
     `));
@@ -156,6 +156,9 @@ var pendingActivePlayer;
 var playerMetadata = {};
 var playerCallbacks = [];
 
+// Playback state communicated via media sessions api
+var playerPlaybackState = "";
+
 var players = [];
 
 var pendingSeekingUpdate = 0;
@@ -178,8 +181,22 @@ addCallback("mpris", "playPause", function () {
     }
 });
 
-// there's no dedicated "stop", simulate it be rewinding and reloading
 addCallback("mpris", "stop", function () {
+    // When available, use the "stop" media sessions action
+    if (playerCallbacks.indexOf("stop") > -1) {
+        executeScript(`
+            function() {
+                try {
+                    ${mediaSessionsClassName}.executeCallback("stop");
+                } catch (e) {
+                    console.warn("Exception executing 'stop' media sessions callback", e);
+                }
+            }
+        `);
+        return;
+    }
+
+    // otherwise since there's no "stop" on the player, simulate it be rewinding and reloading
     if (activePlayer) {
         activePlayer.pause();
         activePlayer.currentTime = 0;
@@ -192,6 +209,7 @@ addCallback("mpris", "stop", function () {
         setTimeout(function() {
             sendMessage("mpris", "stopped");
         }, 1);
+        return;
     }
 });
 
@@ -238,6 +256,7 @@ addCallback("mpris", "setPlaybackRate", function (message) {
 addCallback("mpris", "setVolume", function (message) {
     if (activePlayer) {
         activePlayer.volume = message.volume;
+        activePlayer.muted = (message.volume == 0.0);
     }
 });
 
@@ -297,6 +316,7 @@ function setPlayerActive(player) {
         currentTime: player.currentTime,
         playbackRate: player.playbackRate,
         volume: player.volume,
+        muted: player.muted,
         loop: player.loop,
         metadata: playerMetadata,
         callbacks: playerCallbacks
@@ -304,6 +324,11 @@ function setPlayerActive(player) {
 }
 
 function sendPlayerGone() {
+    var playerIdx = players.indexOf(activePlayer);
+    if (playerIdx > -1) {
+        players.splice(playerIdx, 1);
+    }
+
     activePlayer = undefined;
     pendingActivePlayer = undefined;
     playerMetadata = {};
@@ -345,6 +370,12 @@ function registerPlayer(player) {
     // playlist is now empty or being reloaded, stop player
     // e.g. when using Ajax page navigation and the user nagivated away
     player.addEventListener("emptied", function () {
+        // When the player is emptied but the website tells us it's just "paused"
+        // keep it around (Bug 402324: Soundcloud does this)
+        if (player === activePlayer && playerPlaybackState === "paused") {
+            return;
+        }
+
         // could have its own signal but for compat it's easier just to pretend to have stopped
         sendPlayerInfo(player, "stopped");
     });
@@ -405,11 +436,11 @@ function registerPlayer(player) {
 
     player.addEventListener("volumechange", function () {
         sendPlayerInfo(player, "volumechange", {
-            volume: player.volume
+            volume: player.volume,
+            muted: player.muted
         });
     });
 
-    // TODO remove it again when it goes away
     players.push(player);
 }
 
@@ -464,14 +495,12 @@ function loadMpris() {
                     return;
                 }
 
-                // first check whether the node itself is audio/video
+                // Check whether the node itself or any of its children is a player
+                var players = Array.from(node.querySelectorAll("video,audio"));
                 if (node.matches("video,audio")) {
-                    registerPlayer(node);
-                    return;
+                    players.unshift(node);
                 }
 
-                // if not, check whether any of its children are
-                var players = node.querySelectorAll("video,audio");
                 players.forEach(function (player) {
                     registerPlayer(player);
                 });
@@ -482,16 +511,19 @@ function loadMpris() {
                     return;
                 }
 
+                // Check whether the node itself or any of its children is the current player
+                var players = Array.from(node.querySelectorAll("video,audio"));
                 if (node.matches("video,audio")) {
-                    if (node == activePlayer) {
-                        sendPlayerGone();
-                    }
-                    return;
+                    players.unshift(node);
                 }
 
-                var players = node.querySelectorAll("video,audio");
                 players.forEach(function (player) {
                     if (player == activePlayer) {
+                        // If the player is still in the visible DOM, don't consider it gone
+                        if (document.body.contains(player)) {
+                            return; // continue
+                        }
+
                         sendPlayerGone();
                         return;
                     }
@@ -562,7 +594,7 @@ function loadMediaSessionsShim() {
                 ${mediaSessionsClassName} = function() {};
                 ${mediaSessionsClassName}.transferItem = null;
                 ${mediaSessionsClassName}.callbacks = {};
-                ${mediaSessionsClassName}.metadata = {};
+                ${mediaSessionsClassName}.metadata = null;
                 ${mediaSessionsClassName}.playbackState = "none";
                 ${mediaSessionsClassName}.sendMessage = function(action, payload) {
                     this.transferItem.innerText = JSON.stringify({action: action, payload: payload});
@@ -575,20 +607,51 @@ function loadMediaSessionsShim() {
                     this.callbacks[action]();
                 };
 
-                navigator.mediaSession = {};
+                if (!navigator.mediaSession) {
+                    navigator.mediaSession = {};
+                }
+
+                var noop = function() { };
+
+                var oldSetActionHandler = navigator.mediaSession.setActionHandler || noop;
                 navigator.mediaSession.setActionHandler = function (name, cb) {
+                    // Call the original native implementation
+                    // "call()" is needed as the real setActionHandler is a class member
+                    // and calling it directly is illegal as it lacks the context
+                    // We'll register the callback for ourself after this since it may
+                    // throw for unsupported callback names.
+                    var ret = oldSetActionHandler.call(navigator.mediaSession, name, cb);
+
                     if (cb) {
                         ${mediaSessionsClassName}.callbacks[name] = cb;
                     } else {
                         delete ${mediaSessionsClassName}.callbacks[name];
                     }
                     ${mediaSessionsClassName}.sendMessage("callbacks", Object.keys(${mediaSessionsClassName}.callbacks));
+
+                    return ret;
                 };
+
                 Object.defineProperty(navigator.mediaSession, "metadata", {
                     get: function() { return ${mediaSessionsClassName}.metadata; },
                     set: function(newValue) {
                         ${mediaSessionsClassName}.metadata = newValue;
-                        ${mediaSessionsClassName}.sendMessage("metadata", newValue ? newValue.data : null);
+
+                        // MediaMetadata is not a regular Object so we cannot just JSON.stringify it
+                        var newMetadata = {};
+                        if (newValue) {
+                            var keys = Object.getOwnPropertyNames(Object.getPrototypeOf(newValue));
+
+                            keys.forEach(function (key) {
+                                var value = newValue[key];
+                                if (typeof value === "function") {
+                                    return; // continue
+                                }
+                                newMetadata[key] = newValue[key];
+                            });
+                        }
+
+                        ${mediaSessionsClassName}.sendMessage("metadata", newMetadata);
                     }
                 });
                 Object.defineProperty(navigator.mediaSession, "playbackState", {
@@ -599,9 +662,15 @@ function loadMediaSessionsShim() {
                     }
                 });
 
-                window.MediaMetadata = function (data) {
-                    this.data = data;
-                };
+                if (!window.MediaMetadata) {
+                    window.MediaMetadata = function (data) {
+                        Object.assign(this, data);
+                    };
+                    window.MediaMetadata.prototype.title = "";
+                    window.MediaMetadata.prototype.artist = "";
+                    window.MediaMetadata.prototype.album = "";
+                    window.MediaMetadata.prototype.artwork = [];
+                }
             }
         `);
 
@@ -678,6 +747,7 @@ function loadMediaSessionsShim() {
             } else if (action === "playbackState") {
                 var playbackState = json.payload;
 
+                playerPlaybackState = playbackState;
                 if (activePlayer) {
                     if (playbackState === "playing") {
                         playerPlaying(activePlayer);
